/*
 * gotr_plugin.c
 *
 *  Created on: Mar 22, 2012
 *      Author: lmm
 *      Copied a lot of pidgin demo code
 *      */
#include <glib.h>
#include <gcrypt.h>
#include <pthread.h>

//#include "internal.h"
#include <plugin.h>
#include <cipher.h>
#include <connection.h>
#include <conversation.h>
#include <core.h>
#include <debug.h>
#include <ft.h>
#include <signals.h>
#include <version.h>
#include <status.h>
#include <cmds.h>
#include <sound.h>

#include "gotr_utility.h"

/**
 *  thread packet for gcrypt to use
static struct gcry_thread_cbs gcry_pthread = {
  (GCRY_THREAD_OPTION_USER | (GCRY_THREAD_OPTION_VERSION << 8)),
  NULL, pthread_mutex_init, pthread_mutex_destroy,
  pthread_mutex_lock, pthread_mutex_unlock,
  NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL };
 */

#define GOTR_PLUGIN_ID    "core-lmm-gotr"
#define GOTR_PLUGIN_VERSION	"0.0"
#define GOTR_PLUGIN_WEBSITE "people.cis.ksu.edu/~hongl/gotr/"
#define GOTR_KEYUPDATE_USERCMD "kd"
#define GOTR_KEYUPDATE_HELP "update keys with all members in a chat"

#define TESTRND (100)

PurplePlugin *gotr_plugin_handle;
void *core_handle     ;
void *blist_handle    ;
void *conn_handle     ;
void *conv_handle     ;
void *accounts_handle ;
void *ciphers_handle  ;
void *ft_handle       ;
void *sound_handle    ;
void *notify_handle   ;
void *jabber_handle   ;


/**
 * Used to hold a handle to the commands we register. Holding this handle
 * allows us to unregister the commands at a later time.
 */
static PurpleCmdId keyupdate_command_id;

/**
 * generate a prime from chat info, the func is the same for anyone
static char*
gotr_gen_prime(const PurpleConvChat* chat) {
	gpg_error_t err;
	char* name;
	char salt[8] = {0};
	unsigned char buf[GOTR_PRIMESIZE/8];
	size_t n;
	if(NULL == chat || NULL == chat->conv || NULL == (name = chat->conv->name))
		return NULL;
	n = strlen(chat->conv->title);
	memcpy(salt, chat->conv->title, n);
	if(n < 8)
		memcpy(salt+n, name, n-8);
	err = gcry_kdf_derive(name, strlen(name), GCRY_KDF_ITERSALTED_S2K, 0, salt, 8, 8, GOTR_PRIMESIZE/8, buf);
	if(err)
		return NULL;
	return gotr_printbytes(buf, GOTR_PRIMESIZE/8);
}
 */

// TODO: note that the first param  is to be removed!
/**
 * pack msg according to current state with a partner
 * Note that this func will NOT change states
 * @param 1: chat data
 * @param 2: account name of the partner to send msg to
 * @return: a newly created GByteArray that stores the packed msg or NULL on error
 */
static GByteArray *
gotr_pack_idv(PurpleConversation *conv, gotr_t* chatdata, const char* hername, gcry_error_t* perr)
{
	gotr_keys_t* memkeys = NULL;
	gotr_keys_t* handle = NULL;
	unsigned char iv[GOTR_BLOCKLEN] = {0};
	unsigned char buf[4] = {0};
	size_t nwritten = 0;
	size_t i = 0;
	gcry_error_t err = 0;
	unsigned char* mpibuf = NULL;
	unsigned char* sig_base = NULL;
	unsigned char* sig = NULL;
	GByteArray* text = NULL;
	gotr_state_t state;

	char myname[256] = {0}; // TODO: to be removed
	const char* str;
	char* pz;
	PurpleAccount* myacc;
	if(NULL == conv || NULL == (myacc = purple_conversation_get_account(conv)) || NULL == (str = purple_account_get_username(myacc))) {
		return NULL;
	}
	pz = myname;
	while('@' != *str) {
		*pz++ = *str++;
	}
	// end of to be removed

	while(1) {
		if(NULL == chatdata) {
			err = GOTR_ERR_MEM; gotr_checkerr(err);
		}
		memkeys = gotr_chatdata_getmap(chatdata, hername);
		if(NULL == memkeys) {
			err = GOTR_ERR_MEM; gotr_checkerr(err);
		}
		mpibuf = gcry_malloc(GOTR_MAXPRPKEYLEN);
		if(NULL == mpibuf) {
			err = GOTR_ERR_MEM; gotr_checkerr(err);
		}
		text = g_byte_array_sized_new(GOTR_PLAINTXTLEN);
		if(NULL == text) {
			err = GOTR_ERR_MEM; gotr_checkerr(err);
		}
		break;
	}
	if(err) {
		if(perr)
			*perr = err;
		return NULL;
	}

	g_byte_array_append(text, buf, 4);				// 0. 4-byte length preallocated.
	// the length is calc from head to the end of the msg without paddings and signature

	gcry_create_nonce(iv, GOTR_BLOCKLEN);
	g_byte_array_append(text, iv, GOTR_BLOCKLEN);		// 1. initialization vector, may not useful

	state = memkeys->state;
	switch(state) {
	case GOTR_STATE_PAIR_INIT_S:
	case GOTR_STATE_PAIR_INIT_R : {
		// create a new key_set and a pairwise key for this pairwise session
		gcry_cipher_hd_t hds = NULL;
		unsigned char* hash = NULL;
		while(1) {
			hash = gcry_malloc (gcry_md_get_algo_dlen(GOTR_HASH));
			if(NULL == hash) {
				err = GOTR_ERR_MEM;
				gotr_checkerr(err);
			}

			err = gcry_cipher_open (&hds, GOTR_CIPHER, GCRY_CIPHER_MODE_CBC, GCRY_CIPHER_CBC_MAC);
			gotr_checkerr(err);
			if(NULL == hds) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}

			if(NULL == (handle = memkeys->next)) {// in case of retrans
				handle = memkeys->next = gotr_create_memdata();
				if(NULL == handle) {
					err = GOTR_ERR_MEM; gotr_checkerr(err);
				}
				handle->old = memkeys;
			}

			// gen BD keys
			if(NULL == handle->pubpairr) {
				gotr_genBDkeypair(&handle->prvpairr, &handle->pubpairr, chatdata);
			}

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 2);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			g_byte_array_append(text, (const guint8*) hername, strlen(hername)+1);	// 4. name pretending to be half sig - a workaround TODO: remove this
			g_byte_array_append(text, (const guint8*) myname, strlen(myname)+1);	// 4. name pretending to be half sig - a workaround TODO: remove this

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->pubpairr);	// 5. pubpairr
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten); // there is a `0' appended

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			// TODO: sig is derived from denAKE, now is just hash(myname)
			sig_base = gcry_xmalloc(GOTR_BLOCKLEN + GOTR_BLOCKLEN - 1);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM;
				gotr_checkerr(err);
			}
			sig = (unsigned char*)gotr_ceil((size_t)sig_base);
			gcry_md_hash_buffer (GOTR_HASH, hash, myname, strlen(myname));
			err = gcry_cipher_setkey (hds, hash, GOTR_BLOCKLEN);
			gotr_checkerr(err);
			err = gcry_cipher_encrypt (hds,
					sig,
					GOTR_BLOCKLEN,
					text->data,
					text->len);
			gotr_checkerr(err);
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 6. sig

			gotr_chatdata_addrevmap(chatdata, handle->pubpairr, hername, GOTR_KEY_NEXT_PUBPAIRR);	// attach handle->pubpairr to hername
			break;
		}
		if(NULL != hds)
			gcry_cipher_close(hds);
		if(NULL != hash)
			gcry_free(hash);
	}
	break;
	case GOTR_STATE_PAIR_INIT_G_S : 					// send two BD half keys
	case GOTR_STATE_PAIR_INIT_G_R :
		while(1) {
			handle = memkeys->next;
			if(NULL == handle || NULL == handle->prvpairr || NULL == handle->pubpairr || NULL == handle->pubpairk) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->pairrk) {
				gotr_genBDkey(&handle->pairrk, handle->pubpairk, handle->prvpairr, chatdata->prime);
			}
			for(i = 0; i < 2; i++) {			// gen BD keys
				if(NULL == handle->gr[i]) {
					gotr_genBDkeypair(&handle->r[i], &handle->gr[i], chatdata);
				}
			}

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 2);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->pubpairr);	// 4. pubpairr
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->pubpairk);// 5. pubpairk
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[0]);	// 6. gr[0]
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[1]);// 7. gr[1]
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			sig_base = gotr_sig(handle->pairrk, text->data, text->len, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 8. sig
			gotr_chatdata_addrevmap(chatdata, handle->gr[0], hername, GOTR_KEY_NEXT_GR);
			gotr_chatdata_addrevmap(chatdata, handle->gr[1], hername, GOTR_KEY_NEXT_GR);
			break;
		}
		break;
	case GOTR_STATE_PAIR_INIT_X_S :
	case GOTR_STATE_PAIR_INIT_X_R :
		// comp and send two gadgets
		while(1) {
			handle = memkeys->next;
			if(NULL == handle || NULL == handle->prvpairr || NULL == handle->pubpairr || NULL == handle->pubpairk || NULL == handle->pairrk) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->r[0] || NULL == handle->r[1] || NULL == handle->gr[0] || NULL == handle->gr[1] || NULL == handle->gk[0] || NULL == handle->gk[1]) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->x0[0])
				gotr_gengadget(&handle->x0[0], handle->gr[1], handle->gk[0], handle->r[0], chatdata->prime);
			if(NULL == handle->x0[1])
				gotr_gengadget(&handle->x0[1], handle->gk[1], handle->gr[0], handle->r[1], chatdata->prime);

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 2);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->pubpairr);	// 4. pubpairr
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->pubpairk);//  pubpairk
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->x0[0]);	// 5. x0
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->x0[1]);//
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			sig_base = gotr_sig(handle->pairrk, text->data, text->len, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 8. sig
			break;
		}
		break;
	case GOTR_STATE_PAIR_INIT_V_S :
	case GOTR_STATE_PAIR_UPDATE_V_S :
		while (1) {
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->r[0] || NULL == handle->r[1] || NULL == handle->gr[0] || NULL == handle->gr[1] || NULL == handle->gk[0] || NULL == handle->gk[1]) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->x0[0] || NULL == handle->x0[1] || NULL == handle->x2[0] || NULL == handle->x2[1]) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			err = gotr_calcflakekey(chatdata, handle);
			gotr_checkerr(err);

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 4);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[0]);	// 4.
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gk[0]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gk[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			gcry_create_nonce (iv, GOTR_BLOCKLEN);			// first 4B are nonce, padded to GOTR_BLOCKLEN
			handle->noncer = read_ntohl(iv);
			g_byte_array_append(text, iv, GOTR_BLOCKLEN);				// 5. a plain nonce

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			gotr_printf("pack_state %d, ", state);
			gotr_printmpi("flakek: ", handle->flakek, "\n");

			sig_base = gotr_sig(handle->flakek, text->data, text->len, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 8. sig
			break;
		}
		break;
	case GOTR_STATE_PAIR_INIT_V_R :
	case GOTR_STATE_PAIR_UPDATE_V_R : {
		unsigned char* ebuf_base = NULL;
		unsigned char* ebuf;
		while (1) {
			gcry_cipher_hd_t hde;

			handle = memkeys->next;
			if(NULL == handle) {	// now pubpairrk is the flake key
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->r[0] || NULL == handle->r[1] || NULL == handle->gr[0] || NULL == handle->gr[1] || NULL == handle->gk[0] || NULL == handle->gk[1]) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->x0[0] || NULL == handle->x0[1] || NULL == handle->x2[0] || NULL == handle->x2[1]) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			err = gotr_calcflakekey(chatdata, handle);
			gotr_checkerr(err);

			err = gcry_cipher_open (&hde, GOTR_CIPHER, GCRY_CIPHER_MODE_CBC, 0);
			gotr_checkerr(err);
			if(NULL == hde) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			err = gcry_cipher_setiv (hde, iv, GOTR_BLOCKLEN);
			gotr_checkerr(err);

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 4);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[0]);	// 4.
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gk[0]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gk[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			gotr_printf("pack_state %d, ", state);
			gotr_printmpi("flakek: ", handle->flakek, "\n");

			write_htonl(buf, handle->noncek);
			ebuf_base = gotr_enc(iv, handle->flakek, buf, 4, &ebuf, &nwritten);
			if(NULL == ebuf_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, ebuf, nwritten);				// 5. nonce

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			sig_base = gotr_sig(handle->flakek, text->data, text->len, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 8. sig
			break;
		}
		if(NULL != ebuf_base)
			gcry_free(ebuf_base);
	}
	break;
	case GOTR_STATE_PAIR_UPDATE_G_S :
	case GOTR_STATE_PAIR_UPDATE_G_R :
		// create a new key_set
		while(1) {
			handle = memkeys->next;
			if(NULL == handle) {// in case of retrans
				handle = memkeys->next = gotr_create_memdata();
				if(NULL == handle) {
					err = GOTR_ERR_MEM; gotr_checkerr(err);
				}
				handle->old = memkeys;
			}
			if(NULL == memkeys->flakek) {	// flake key
				err = GOTR_ERR_STAT;
				gotr_checkerr(err);
			}

			for(i = 0; i < 2; i++) {			// gen BD keys
				if(NULL == handle->gr[i]) {
					gotr_genBDkeypair(&handle->r[i], &handle->gr[i], chatdata);
				}
			}

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 4);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gr[0]);	// 4.
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gr[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gk[0]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gk[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[0]);	// 5. gr[0]
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->gr[1]);// gr[1]
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			sig_base = gotr_sig(memkeys->flakek, text->data, text->len, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 8. sig

			gotr_chatdata_addrevmap(chatdata, handle->gr[0], hername, GOTR_KEY_NEXT_GR);
			gotr_chatdata_addrevmap(chatdata, handle->gr[1], hername, GOTR_KEY_NEXT_GR);
			break;
		}
		break;
	case GOTR_STATE_PAIR_UPDATE_X_S :
	case GOTR_STATE_PAIR_UPDATE_X_R :
		// comp and send two gadgets
		while(1) {
			handle = memkeys->next;
			if(NULL == handle || NULL == memkeys->flakek) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->r[0] || NULL == handle->r[1] || NULL == handle->gr[0] || NULL == handle->gr[1] || NULL == handle->gk[0] || NULL == handle->gk[1]) {
				err = GOTR_ERR_STAT;	 // indicates wrong status
				gotr_checkerr(err);
			}
			if(NULL == handle->x0[0])
				gotr_gengadget(&handle->x0[0], handle->gr[1], handle->gk[0], handle->r[0], chatdata->prime);
			if(NULL == handle->x0[1])
				gotr_gengadget(&handle->x0[1], handle->gk[1], handle->gr[0], handle->r[1], chatdata->prime);

			write_htons(buf, memkeys->state);	// 2. op
			g_byte_array_append(text, buf, 2);

			write_htonl(buf, 4);	// 3. # of pub BD keys following
			g_byte_array_append(text, buf, 4);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gr[0]);	// 4.
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gr[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gk[0]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memkeys->gk[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->x0[0]);	// 5. x0
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, handle->x0[1]);//
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);

			write_htonl(text->data, text->len);					// write len to the head

			nwritten = gotr_ceil(text->len);	// padding
			if(nwritten > text->len) {
				gcry_create_nonce (iv, GOTR_BLOCKLEN);
				g_byte_array_append(text, iv, nwritten - text->len);
			}

			sig_base = gotr_sig(memkeys->flakek, text->data, text->len, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, sig, GOTR_BLOCKLEN); // 8. sig
			break;
		}
		break;
	default :
		err = GOTR_ERR_STATE; break;
	}

	gotr_printf("~~~~~~~~~~~~~~\nstate: %d, err = %d\n", memkeys->state, err);
	gotr_printflakekey(memkeys);
	gotr_printf("~~~~~~\n");
	gotr_printflakekey(handle);
	gotr_printf("~~~~~~~~~~~~~~\n");

	if(NULL != mpibuf)
		gcry_free(mpibuf);
	if(NULL != sig_base)
		gcry_free(sig_base);
	if(perr)
		*perr = err;
	if(err) {
		if(text)
			g_byte_array_free(text, TRUE);
		return NULL;
	}
	else {
		return text;
	}
}

/**
 * pack msg according to current state with a partner
 * Note that this func will NOT change states
 * @param 1: chat data
 * @param 2: msg, can be NULL
 * @return: a newly created GByteArray that stores the packed msg or NULL on error
 */
static GByteArray *
gotr_pack_brdcast_msg(gotr_t* chatdata, const char* data) {
	unsigned char iv[GOTR_BLOCKLEN];
	unsigned char buf[4];
	size_t nwritten;
	gcry_error_t err = 0;
	unsigned char* mpibuf = NULL;
	GByteArray* text = NULL;
	GList* circle = NULL;
	size_t n = 0;
	gcry_mpi_t circlekey = NULL;

	while(1) {
		gotr_checkerr(NULL == chatdata);

		circle = gotr_chatdata_getvalidkeyls(chatdata, &circlekey);
		gotr_checkerr(NULL == circle || NULL == circlekey);
		gotr_printmpi("enc: ", circlekey, "\n");
		n = g_list_length (circle);
		gotr_checkerr(0 == n);

		gcry_create_nonce(iv, GOTR_BLOCKLEN);

		mpibuf = gcry_malloc(GOTR_MAXPRPKEYLEN);
		gotr_checkerr(NULL == mpibuf);

		text = g_byte_array_sized_new(GOTR_PLAINTXTLEN);
		gotr_checkerr(NULL == text);
		break;
	}
	if(NULL == text)
		return NULL;

	g_byte_array_append(text, buf, 4);				// 0. 4-byte length preallocated.
	// the length is calc from head to the end of the msg without paddings and signature

	g_byte_array_append(text, iv, GOTR_BLOCKLEN);	// 1. initialization vector, prob not useful

	write_htons(buf, GOTR_STATE_MSG);							// 2. op
	g_byte_array_append(text, buf, 2);

	write_htonl(buf, 8*n);
	g_byte_array_append(text, buf, 4);				// 3. # of pub BD keys preallocated. g_1, ... g_{4n}, X_1, ..., X_{4n}

	while(1) {
		GList* ls = circle;
		gotr_keys_t* memdata;
		unsigned char* cbuf;
		unsigned char* cbufbase;
		while(NULL != ls) {
			memdata = (gotr_keys_t*)ls->data;
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->gr[0]);	// 4. g_1, ... g_{4n}
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->gk[0]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->gk[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->gr[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			ls = ls->next;
		}
		ls = circle;
		while(NULL != ls) {
			memdata = (gotr_keys_t*)ls->data;
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->x1[0]);	// 4. X_1, ..., X_{4n}
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->x2[0]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->x2[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			err = gcry_mpi_print(GCRYMPI_FMT_HEX, mpibuf, GOTR_MAXPRPKEYLEN, &nwritten, memdata->x1[1]);
			gotr_checkerr(err);
			g_byte_array_append(text, mpibuf, nwritten);
			ls = ls->next;
		}

		if(NULL != data) {
			write_htonl(text->data, text->len + strlen(data));					// write len to the head
			cbufbase = gotr_enc(iv, circlekey, (const unsigned char*)data, strlen(data), &cbuf, &nwritten); // 5. enc
			if(NULL == cbufbase) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			g_byte_array_append(text, cbuf, nwritten);
			gcry_free(cbufbase);
		}
		else {
			write_htonl(text->data, text->len);					// write len to the head
		}

		nwritten = gotr_ceil(text->len);	// padding
		if(nwritten > text->len) {
			gcry_create_nonce (iv, GOTR_BLOCKLEN);
			g_byte_array_append(text, iv, nwritten - text->len);
		}

		cbufbase = gotr_sig(circlekey, text->data, text->len, &cbuf);	// 6. sig
		if(NULL == cbufbase) {
			err = GOTR_ERR_CRYPT;
			gotr_checkerr(err);
		}
		g_byte_array_append(text, cbuf, GOTR_BLOCKLEN);
		gcry_free(cbufbase);
		break;
	}

	if(NULL != circle)
		g_list_free(circle);
	if(NULL != mpibuf)
		gcry_free(mpibuf);
	if(NULL != circlekey)
		gcry_mpi_release(circlekey);
	if(err) {
		if(text)
			g_byte_array_free(text, TRUE);
		return NULL;
	}
	else {
		return text;
	}
}

static void
gotr_send_chat_msg(PurpleConversation *conv, GByteArray* payload) {
	char *str;
	if(NULL == conv || NULL == payload)
		return;
	str = (char*)malloc(2*payload->len+2);
	*str = GOTR_ESC;
	gotr_postproc_msg(payload->data, payload->len, str+1);
	// send init/update msg to this member
//	purple_signal_emit(purple_conversations_get_handle(), "sending-chat-msg",
//			purple_conversation_get_account(conv), &str, purple_conv_chat_get_id(purple_conversation_get_chat_data(conv)));

	purple_conv_chat_send(purple_conversation_get_chat_data(conv), str);
}

// TODO: conv to be removed
/**
 * unpack msg according to current state with a partner
 *  Note that this func will NOT change states
 * 	derive sender's identity from the piggyback pubkeys
 * @param 1: chat data
 * @param 2: msg, cant be NULL
 * @param 3: msg len
 * @param 4: a newly created derived sender id
 * @param 5: error. 0 if no error
 * @return: a newly created GByteArray that stores the unpacked msg or NULL on error or if the msg doesn't have payload
 */
static GByteArray *
gotr_unpack_msg(PurpleConversation *conv, gotr_t* chatdata,  const unsigned char* data, size_t datalen, const char** hername, gcry_error_t* perr) {
	gotr_keys_t* memkeys = NULL;
	gotr_keys_t* handle = NULL;
	gcry_error_t err = 0;
	unsigned char iv[GOTR_BLOCKLEN];
	gotr_state_t op;
	size_t numPubkeys;
	size_t len;
	size_t offset = 0;
	unsigned char* sig_base = NULL;
	unsigned char* sig = NULL;
	GByteArray* text = NULL;
	size_t nwritten = 0;

	char myname[256] = {0}; // TODO: to be removed
	const char* str = NULL;
	char* pz = NULL;
	PurpleAccount* myacc = NULL;
	if(NULL == conv || NULL == (myacc = purple_conversation_get_account(conv)) || NULL == (str = purple_account_get_username(myacc))) {
		return NULL;
	}
	pz = myname;
	while(*str && '@' != *str) {
		*pz++ = *str++;
	}
	*pz = 0;
	// end of to be removed

	if(NULL == chatdata || NULL == data
			|| 0 == datalen
			|| NULL == hername) {
		if(NULL != perr)
			*perr = GOTR_ERR_PRM;
		return NULL;
	}

	while(1) {
		len = read_ntohl(data+offset);							// 0. len
		if(gotr_ceil(len) + GOTR_BLOCKLEN > datalen) {
			err = GOTR_ERR_PRM;
			gotr_checkerr(err);
		}
		offset += 4;

		memcpy(iv, data+offset, GOTR_BLOCKLEN);			//  1. initialization vector
		offset += GOTR_BLOCKLEN;

		op = read_ntohs(data+offset);	// 2. op
		offset += 2;

		numPubkeys = read_ntohl(data+offset);			// 3. # of pub BD keys following
		offset += 4;
		break;
	}
	if(err) {
		*perr = err;
		return NULL;
	}

	switch(op) {
	case GOTR_STATE_PAIR_INIT_S: {
		gcry_cipher_hd_t hds = NULL;
		unsigned char* hash = NULL;
		while(1) {
			if(strcmp(myname, (char*)data+offset) != 0) {
				err = GOTR_ERR_OTHER; gotr_checkerr(err);
			}
			offset += strlen((char*)data+offset) + 1;

			sig_base = gcry_xmalloc(GOTR_BLOCKLEN + GOTR_BLOCKLEN - 1);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM;
				gotr_checkerr(err);
			}
			sig = (unsigned char*)gotr_ceil((size_t)sig_base);

			err = gcry_cipher_open (&hds, GOTR_CIPHER, GCRY_CIPHER_MODE_CBC, GCRY_CIPHER_CBC_MAC);
			gotr_checkerr(err);
			if(NULL == hds) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			hash = gcry_malloc (gcry_md_get_algo_dlen(GOTR_HASH));
			if(NULL == hash) {
				err = GOTR_ERR_MEM;
				gotr_checkerr(err);
			}
			gcry_md_hash_buffer (GOTR_HASH, hash, data+offset, strlen((char*)data+offset));	// 4. name pretending to be half sig - a workaround TODO: remove this
			err = gcry_cipher_setkey (hds, hash, GOTR_BLOCKLEN);
			gotr_checkerr(err);
			err = gcry_cipher_encrypt (hds,
					sig,
					GOTR_BLOCKLEN,
					data,
					gotr_ceil(len));
			gotr_checkerr(err);
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}

			if(gotr_chatdata_lookupname(chatdata, (char*)data+offset)) { // such name already recorded
				memkeys = gotr_chatdata_getmap(chatdata, (char*)data+offset);
				if(NULL == memkeys || GOTR_STATE_IDLE != memkeys->state || NULL == (handle = memkeys->next)) {
					err = GOTR_ERR_STAT; gotr_checkerr(err);
				}
			}
			else { // receive a init request from a newcomer
				memkeys = gotr_create_memdata();
				if(NULL == memkeys) {
					err = GOTR_ERR_MEM; gotr_checkerr(err);
				}
				gotr_chatdata_addmap(chatdata, (char*)data+offset, memkeys);

				handle = memkeys->next = gotr_create_memdata();
				if(NULL == handle) {
					err = GOTR_ERR_MEM; gotr_checkerr(err);
				}
				handle->old = memkeys;
			}
			*hername = gotr_chatdata_getstaticname(chatdata, (char*)data+offset);
			// get half BD key and store it to handle->pubpairk
			offset += strlen((char*)data+offset) + 1;
			err = gcry_mpi_scan(&handle->pubpairk, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			// delay calc pairwise key until send'g
			gotr_chatdata_addrevmap(chatdata, handle->pubpairk, *hername, GOTR_KEY_NEXT_PUBPAIRK);	// attach handle->pubpairk to hername
			break;
		}
		if(NULL != hds)
			gcry_cipher_close(hds);
		if(NULL != hash)
			gcry_free(hash);
	}
	break;
	case GOTR_STATE_PAIR_INIT_R: {
		gcry_cipher_hd_t hds = NULL;
		unsigned char* hash = NULL;
		while(1) {
			if(strcmp(myname, (char*)data+offset) != 0) {
				err = GOTR_ERR_OTHER; gotr_checkerr(err);
			}
			offset += strlen((char*)data+offset) + 1;

			sig_base = gcry_xmalloc(GOTR_BLOCKLEN + GOTR_BLOCKLEN - 1);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM;
				gotr_checkerr(err);
			}
			sig = (unsigned char*)gotr_ceil((size_t)sig_base);

			err = gcry_cipher_open (&hds, GOTR_CIPHER, GCRY_CIPHER_MODE_CBC, GCRY_CIPHER_CBC_MAC);
			gotr_checkerr(err);
			if(NULL == hds) {
				err = GOTR_ERR_CRYPT;
				gotr_checkerr(err);
			}
			hash = gcry_malloc (gcry_md_get_algo_dlen(GOTR_HASH));
			if(NULL == hash) {
				err = GOTR_ERR_MEM;
				gotr_checkerr(err);
			}
			gcry_md_hash_buffer (GOTR_HASH, hash, data+offset, strlen((char*)data+offset));	// 4. name pretending to be half sig - a workaround TODO: remove this
			err = gcry_cipher_setkey (hds, hash, GOTR_BLOCKLEN);
			gotr_checkerr(err);
			err = gcry_cipher_encrypt (hds,
					sig,
					GOTR_BLOCKLEN,
					data,
					gotr_ceil(len));
			gotr_checkerr(err);
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}

			if(gotr_chatdata_lookupname(chatdata, (char*)data+offset)) { // such name already recorded
				memkeys = gotr_chatdata_getmap(chatdata, (char*)data+offset);
				if(NULL == memkeys || GOTR_STATE_PAIR_INIT_S != memkeys->state || NULL == (handle = memkeys->next)) {
					err = GOTR_ERR_STAT; gotr_checkerr(err);
				}
			}
			else {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			*hername = gotr_chatdata_getstaticname(chatdata, (char*)data+offset);
			// get half BD key and store it to handle->pubpairk
			offset += strlen((char*)data+offset) + 1;
			err = gcry_mpi_scan(&handle->pubpairk, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			// delay calc pairwise key until send'g
			gotr_chatdata_addrevmap(chatdata, handle->pubpairk, *hername, GOTR_KEY_NEXT_PUBPAIRK);	// attach handle->pubpairk to hername
			break;
		}
		if(NULL != hds)
			gcry_cipher_close(hds);
		if(NULL != hash)
			gcry_free(hash);
	}
	break;
	case GOTR_STATE_PAIR_INIT_G_S :
	case GOTR_STATE_PAIR_INIT_G_R : {
		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_NEXT_PUBPAIRR == id0->type) {
				t = GOTR_KEY_NEXT_PUBPAIRK;
			}
			else if(GOTR_KEY_NEXT_PUBPAIRK == id0->type) {
				t = GOTR_KEY_NEXT_PUBPAIRR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(op - 1 != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(NULL == handle->pairrk) {
				gotr_genBDkey(&handle->pairrk, handle->pubpairk, handle->prvpairr, chatdata->prime);
			}
			sig_base = gotr_sig(handle->pairrk, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}
			*hername = id1->name;

			err = gcry_mpi_scan(&handle->gk[0], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;

			err = gcry_mpi_scan(&handle->gk[1], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);

			gotr_chatdata_addrevmap(chatdata, handle->gk[0], id1->name, GOTR_KEY_NEXT_GK);
			gotr_chatdata_addrevmap(chatdata, handle->gk[1], id1->name, GOTR_KEY_NEXT_GK);
			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_PAIR_INIT_X_S:
	case GOTR_STATE_PAIR_INIT_X_R: {
		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_NEXT_PUBPAIRR == id0->type) {
				t = GOTR_KEY_NEXT_PUBPAIRK;
			}
			else if(GOTR_KEY_NEXT_PUBPAIRK == id0->type) {
				t = GOTR_KEY_NEXT_PUBPAIRR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(op - 1 != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(NULL == handle->pairrk) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			sig_base = gotr_sig(handle->pairrk, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}
			*hername = id1->name;

			err = gcry_mpi_scan(&handle->x2[0], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;

			err = gcry_mpi_scan(&handle->x2[1], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);

			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_PAIR_INIT_V_S :
	case GOTR_STATE_PAIR_UPDATE_V_S :{
		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_NEXT_GR == id0->type) {
				t = GOTR_KEY_NEXT_GK;
			}
			else if(GOTR_KEY_NEXT_GK == id0->type) {
				t = GOTR_KEY_NEXT_GR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(op - 1 != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			err = gotr_calcflakekey(chatdata, handle);
			gotr_checkerr(err);

			gotr_printf("unpack_state %d, ", op);
			gotr_printmpi("flakek: ", handle->flakek, "\n");

			sig_base = gotr_sig(handle->flakek, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG; gotr_checkerr(err);
			}
			*hername = id1->name;

			handle->noncek = read_ntohl(data+offset);
			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_PAIR_INIT_V_R :
	case GOTR_STATE_PAIR_UPDATE_V_R :{
		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		unsigned char* buf = NULL;
		unsigned char* buf_base = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		uint32_t nonce ;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_NEXT_GR == id0->type) {
				t = GOTR_KEY_NEXT_GK;
			}
			else if(GOTR_KEY_NEXT_GK == id0->type) {
				t = GOTR_KEY_NEXT_GR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(op - 1 != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			err = gotr_calcflakekey(chatdata, handle);
			gotr_checkerr(err);

			gotr_printf("unpack_state %d, ", op);
			gotr_printmpi("flakek: ", handle->flakek, "\n");

			sig_base = gotr_sig(handle->flakek, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG; gotr_checkerr(err);
			}
			*hername = id1->name;

			buf_base = gotr_dec(iv, handle->flakek, data+offset, GOTR_BLOCKLEN, &buf, &nwritten);
			if(NULL == buf_base) {
				err = GOTR_ERR_CRYPT; gotr_checkerr(err);
			}
			nonce = read_ntohl(buf);
			if(nonce != handle->noncer){
				err = GOTR_ERR_VRF; gotr_checkerr(err);
			}
			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(NULL != buf_base)
			gcry_free(buf_base);
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_PAIR_UPDATE_G_S : {

		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_GR == id0->type) {
				t = GOTR_KEY_GK;
			}
			else if(GOTR_KEY_GK == id0->type) {
				t = GOTR_KEY_GR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(GOTR_STATE_MSG != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(NULL == memkeys->flakek) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				handle = memkeys->next = gotr_create_memdata();
				if(NULL == handle) {
					err = GOTR_ERR_MEM; gotr_checkerr(err);
				}
				handle->old = memkeys;
			}
			sig_base = gotr_sig(memkeys->flakek, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}
			*hername = id1->name;

			err = gcry_mpi_scan(&handle->gk[0], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;

			err = gcry_mpi_scan(&handle->gk[1], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);

			gotr_chatdata_addrevmap(chatdata, handle->gk[0], id1->name, GOTR_KEY_NEXT_GK);
			gotr_chatdata_addrevmap(chatdata, handle->gk[1], id1->name, GOTR_KEY_NEXT_GK);
			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_PAIR_UPDATE_G_R : {
		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_GR == id0->type) {
				t = GOTR_KEY_GK;
			}
			else if(GOTR_KEY_GK == id0->type) {
				t = GOTR_KEY_GR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(op - 1 != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(NULL == memkeys->flakek) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			sig_base = gotr_sig(memkeys->flakek, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}
			*hername = id1->name;

			err = gcry_mpi_scan(&handle->gk[0], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;

			err = gcry_mpi_scan(&handle->gk[1], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);

			gotr_chatdata_addrevmap(chatdata, handle->gk[0], id1->name, GOTR_KEY_NEXT_GK);
			gotr_chatdata_addrevmap(chatdata, handle->gk[1], id1->name, GOTR_KEY_NEXT_GK);
			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_PAIR_UPDATE_X_S:
	case GOTR_STATE_PAIR_UPDATE_X_R:{
		GList* ls = NULL;
		derived_id_t* id0 = NULL;
		derived_id_t* id1 = NULL;
		gcry_mpi_t k = NULL;
		gotr_keytype_t t;
		while(1) {
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id0 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(NULL == id0) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(GOTR_KEY_GR == id0->type) {
				t = GOTR_KEY_GK;
			}
			else if(GOTR_KEY_GK == id0->type) {
				t = GOTR_KEY_GR;
			}
			else {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;
			ls = g_list_prepend(ls, k);
			id1 = gotr_chatdata_deriveid(chatdata, ls);
			gotr_list_free_full(ls, gotr_free_mpi);
			ls = NULL;
			k  = NULL;
			if(0 != gotr_compare_names(id0->name, id1->name)) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			if(t != id1->type){
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}

			memkeys = gotr_chatdata_getmap(chatdata, id1->name);
			if(NULL == memkeys) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(op - 1 != memkeys->state) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			handle = memkeys->next;
			if(NULL == handle) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			if(NULL == memkeys->flakek) {
				err = GOTR_ERR_STAT; gotr_checkerr(err);
			}
			sig_base = gotr_sig(memkeys->flakek, data, gotr_ceil(len), &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + gotr_ceil(len), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}
			*hername = id1->name;

			err = gcry_mpi_scan(&handle->x2[0], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);
			nwritten = strlen((char*)data+offset) + 1;
			offset += nwritten;

			err = gcry_mpi_scan(&handle->x2[1], GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
			gotr_checkerr(err);

			break;
		}
		if(NULL != ls)
			gotr_list_free_full(ls, gotr_free_mpi);
		if(NULL != id0) {
			free(id0);
		}
		if(NULL != id1) {
			free(id1);
		}
		if(k) {
			gcry_mpi_release(k);
		}
	}
	break;
	case GOTR_STATE_MSG: {
		GList* lsk = NULL;		// list of pub BD keys
		GList* lsg = NULL;		// list of gadgets
		gcry_mpi_t k = NULL;
		unsigned char* buf_base = NULL;
		unsigned char* buf = NULL;
		size_t i;
		struct timeval tmvfirst;
		gettimeofday(&tmvfirst, NULL);
		while(1) {
			if(0 != numPubkeys%8) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			for(i = 0; i < numPubkeys/2; i++) {
				err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
				gotr_checkerr(err);
				nwritten = strlen((char*)data+offset) + 1;
				offset += nwritten;
				lsk = g_list_prepend(lsk, k);
			}
			gotr_checkerr(err);
			lsk = g_list_reverse(lsk);
			for(; i < numPubkeys; i++) {
				err = gcry_mpi_scan(&k, GCRYMPI_FMT_HEX, data+offset, 0, &nwritten);
				gotr_checkerr(err);
				nwritten = strlen((char*)data+offset) + 1;
				offset += nwritten;
				lsg = g_list_prepend(lsg, k);
			}
			gotr_checkerr(err);
			lsg = g_list_reverse(lsg);
			k = gotr_chatdata_derivecirclek(chatdata, lsk, lsg, hername);
			if(NULL == k) {
				err = GOTR_ERR_PRM; gotr_checkerr(err);
			}
			gotr_printmpi("dec: ", k, "\n");

			sig_base = gotr_sig(k, data, datalen - GOTR_BLOCKLEN, &sig);
			if(NULL == sig_base) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			if(0 != memcmp(sig, data + (datalen - GOTR_BLOCKLEN), GOTR_BLOCKLEN)) {		// test sig first
				err = GOTR_ERR_SIG;
				gotr_checkerr(err);
			}

			buf_base = gotr_dec(iv, k, data+offset, gotr_ceil(len-offset), &buf, &nwritten);
			if(NULL == buf_base) {
				err = GOTR_ERR_CRYPT; gotr_checkerr(err);
			}
			text = g_byte_array_new();
			if(NULL == text) {
				err = GOTR_ERR_MEM; gotr_checkerr(err);
			}
			g_byte_array_append(text, buf, len-offset);
			break;
		}
		if(NULL != lsk)
			gotr_list_free_full(lsk, gotr_free_mpi);
		if(NULL != lsg)
			gotr_list_free_full(lsg, gotr_free_mpi);
		if(NULL != k) {
			gcry_mpi_release(k);
		}
		if(NULL != buf_base) {
			gcry_free(buf_base);
		}
	}
		break;
	default: err = GOTR_ERR_CMD ; break;
	}

	gotr_printflakekey(handle);

	if(NULL != sig_base)
		gcry_free(sig_base);
	if(perr)
		*perr = err;
	if(err) {
		return NULL;
	}
	else {
		return text;
	}
}

/**
 * The callback function for our /keyupdate command. This function simply pops up
 * a notification with the word entered as the argument to the command
 *
 * @param 1: conv The conversation that the command occurred in
 * @param 2: cmd The exact command that was entered
 * @param 3: args The args that were passed with the command
 * @param 4: error ?Any errors that occurred?
 * @param 5: data Any special user-defined data that was assigned during
 *             cmd_register
 */
static PurpleCmdRet
gotr_keyupdate_cb(PurpleConversation *conv, const gchar *cmd, gchar **args, gchar **error, void *data)
{
	GTree* toptable = (GTree*)data;
	PurpleConvChat* chat;
	gotr_t* chatdata;
	GHashTable* keytable;
	gotr_keys_t* memkeys;
	const char *name;
	GList* users;
	GByteArray * msg;
	gcry_error_t err = 0;
//	const char* myname;
//	PurpleAccount* myacc;

	purple_debug_misc("gotr", "command: key-update (%s)\n",
			purple_conversation_get_name(conv));

	if(NULL == toptable)	return PURPLE_CMD_RET_FAILED;

	chat = purple_conversation_get_chat_data(conv);
	if(NULL == chat)		return PURPLE_CMD_RET_FAILED; // not a chat

	chatdata = g_tree_lookup (toptable, GINT_TO_POINTER(chat->id));
	if(NULL == chatdata)	return PURPLE_CMD_RET_FAILED;
	if(NULL == (keytable = chatdata->table))	return PURPLE_CMD_RET_FAILED;

//	myacc = purple_conversation_get_account(conv);
//	if(NULL == myacc)	return PURPLE_CMD_RET_FAILED;
//	myname = purple_account_get_username(myacc);
//	if(NULL == myname)	return PURPLE_CMD_RET_FAILED;

	users = purple_conv_chat_get_users(chat);
	for(; users; users = users->next) {
		name = purple_conv_chat_cb_get_name(users->data);
		if(strcmp(name, chat->nick) == 0)
			continue;
		memkeys = gotr_chatdata_getmap(chatdata, name);
		if(NULL == memkeys) {
			memkeys = gotr_create_memdata();
			// TODO: get static pub keys for this member and stores it to memkeys->pubkey
			memkeys->state = GOTR_STATE_PAIR_INIT_S; // TODO: should be denAKE
			gotr_chatdata_addmap(chatdata, name, memkeys);

			msg = gotr_pack_idv(conv, chatdata, name, &err);
			if(!err) {
				gotr_send_chat_msg(conv, msg);
			}
			else {
				gotr_chatdata_removemap(chatdata, name);
			}
		}
		else if (GOTR_STATE_MSG == memkeys->state) {
			memkeys->state = GOTR_STATE_PAIR_UPDATE_G_S;
			msg = gotr_pack_idv(conv, chatdata, name, &err);
			if(!err) {
				gotr_send_chat_msg(conv, msg);
			}
			else {
				memkeys->state = GOTR_STATE_MSG;
			}
		}
	}
	return PURPLE_CMD_RET_OK;
}

/**
 * Parameters:
    	account 	The account the message is being sent on.
    	message 	A pointer to the message that will be sent.
    	id 	The ID of the chat.
 */
static gboolean
sending_chat_msg_cb(PurpleAccount *account, char **buffer, int id, void *data)
{
	GTree* toptable = (GTree*)data;
	gotr_t* chatdata = NULL;
	char* target = NULL;
	GByteArray * ciphertext;
	gboolean res = FALSE;

	while(1) {
		gotr_checkerr(NULL == toptable);

		chatdata = g_tree_lookup (toptable, GINT_TO_POINTER(id));
		gotr_checkerr(NULL == chatdata);

		if(GOTR_ESC == **buffer) {
			target = g_strdup(*buffer + 1);
		}
		else {
			ciphertext = gotr_pack_brdcast_msg(chatdata, *buffer);

			if(NULL == ciphertext) {
				// TODO: unenc data: need prompt
				target = g_strdup(*buffer);
			}
			else {
				target = gotr_postproc_msg(ciphertext->data, ciphertext->len, NULL);
			}

		}
		res = FALSE;
		break;
	}
	free(*buffer);
	*buffer = target;
	return res;
}

static gboolean
receiving_chat_msg_cb(PurpleAccount *account, char **sender, char **buffer,
		PurpleConversation *conv, PurpleMessageFlags *flags, void *data)
{
	static int count = 0;// for performance msr

	GTree* toptable = (GTree*)data;
	PurpleConvChat* chat = NULL;
	gotr_t* chatdata = NULL;
	size_t pn = 0;
	gcry_error_t err = 0;
	unsigned char* rawdata = NULL;
	GByteArray* payload = NULL;
	const char* derivsender = NULL;
	gotr_keys_t* memdata = NULL;
	char prompt[GOTR_PROMPTLEN+1]={0};
	gboolean stop = TRUE;

	purple_debug_misc("gotr",
			"============================\nreceiving-chat-msg (%s, %s, %s, %s, %d)\n",
			purple_account_get_username(account), *sender, *buffer,
			purple_conversation_get_name(conv), *flags);

	if(NULL == toptable)	return FALSE; // other chats

	chat = purple_conversation_get_chat_data(conv);
	if(NULL == chat)		return FALSE;  // other chats

	while(1) {
		rawdata = gotr_preproc_msg(*buffer, NULL, &pn); // trans *buffer into binary data
		if(NULL == rawdata) {
			snprintf(prompt, GOTR_PROMPTLEN, "receiving-chat-msg errs: not valid encrypted data of length %lu\n", pn);
			err = GOTR_ERR_PRM;
			stop = FALSE;
			gotr_checkerr(err);
		}
		if(pn <= GOTR_BLOCKLEN || 0 != (pn % GOTR_BLOCKLEN)) {
			snprintf(prompt, GOTR_PROMPTLEN, "receiving-chat-msg errs: not valid encrypted data of length %lu\n", pn);
			err = GOTR_ERR_PRM;
			stop = FALSE;
			gotr_checkerr(err);
		}

		free(*buffer);
		*buffer = NULL;
		chatdata = g_tree_lookup (toptable, GINT_TO_POINTER(chat->id));
		if(NULL == chatdata)	{
			snprintf(prompt, GOTR_PROMPTLEN, "failed to lookup data for  chat %d\n", chat->id);
			err = GOTR_ERR_STAT;
			gotr_checkerr(err); // other chats
		}

		payload = gotr_unpack_msg(conv, chatdata, rawdata, pn, &derivsender, &err);	// unpack msg
		if(err) {
			snprintf(prompt, GOTR_PROMPTLEN, "receiving-chat-msg errs: unpack msg errs with err_code %d\n", err);
			err = GOTR_ERR_PRM;
			gotr_checkerr(err);
		}
		if(NULL == derivsender) {
			snprintf(prompt, GOTR_PROMPTLEN, "receiving-chat-msg errs: failed to derive sender's id\n");
			err = GOTR_ERR_STAT;
			gotr_checkerr(err);
		}

		memdata = gotr_chatdata_getmap(chatdata, derivsender);
		if(NULL == memdata) {
			snprintf(prompt, GOTR_PROMPTLEN, "failed to lookup mem data for %s\n", derivsender);
			err = GOTR_ERR_STAT;
			gotr_checkerr(err);
		}
		snprintf(prompt, GOTR_PROMPTLEN, "state == %d\n", memdata->state);
		switch(memdata->state) {	// change state and send msg back
		case GOTR_STATE_IDLE:
		case GOTR_STATE_PAIR_INIT_S:
		case GOTR_STATE_PAIR_INIT_R:
		case GOTR_STATE_PAIR_INIT_G_S:
		case GOTR_STATE_PAIR_INIT_G_R:
		case GOTR_STATE_PAIR_INIT_X_S:
		case GOTR_STATE_PAIR_UPDATE_G_S:
		case GOTR_STATE_PAIR_UPDATE_G_R:
		case GOTR_STATE_PAIR_UPDATE_X_S:
			memdata->state += 2;
			if(NULL != payload)
				g_byte_array_free(payload, TRUE);
			payload = gotr_pack_idv(conv, chatdata, derivsender, &err);
			if(err) {
				snprintf(prompt, GOTR_PROMPTLEN, "failed to pack reply msg %d for %s with err %d\n", memdata->state, derivsender, err);
				memdata->state -= 2;
			}
			else {
				gotr_send_chat_msg(conv, payload);
			}
			break;
		case GOTR_STATE_PAIR_INIT_X_R:
		case GOTR_STATE_PAIR_UPDATE_X_R:
			memdata->state += 2;
			if(NULL != payload)
				g_byte_array_free(payload, TRUE);
			payload = gotr_pack_idv(conv, chatdata, derivsender, &err);
			if(err) {
				snprintf(prompt, GOTR_PROMPTLEN, "failed to pack reply msg %d for %s with err %d\n", memdata->state, derivsender, err);
				memdata->state -= 2;
			}
			else {
				gotr_send_chat_msg(conv, payload);
				memdata->next->state = GOTR_STATE_MSG;
				memdata->state = GOTR_STATE_MSG;
				gotr_chatdata_updatemap(chatdata, derivsender);
			}
			break;
		case GOTR_STATE_PAIR_INIT_V_S:
		case GOTR_STATE_PAIR_UPDATE_V_S:
			memdata->next->state = GOTR_STATE_MSG;
			memdata->state = GOTR_STATE_MSG;
			gotr_chatdata_updatemap(chatdata, derivsender);
			break;
		case GOTR_STATE_MSG:
			if(NULL == payload) {
				memdata->state = GOTR_STATE_PAIR_UPDATE_G_R;
				payload = gotr_pack_idv(conv, chatdata, derivsender, &err);
				if(err) {
					snprintf(prompt, GOTR_PROMPTLEN, "failed to pack reply msg %d for %s with err %d\n", memdata->state, derivsender, err);
					memdata->state = GOTR_STATE_MSG;
				}
				else {
					gotr_send_chat_msg(conv, payload);
				}
			}
			else {
				pn = strlen(derivsender) + payload->len + 10;
				*buffer = (char*)malloc(pn);
				if(NULL == *buffer) {
					err = GOTR_ERR_MEM; gotr_checkerr(err);
				}
				snprintf(*buffer, pn, "%s: ", derivsender);
				pn = strlen(*buffer);
				memcpy(*buffer + pn, payload->data, payload->len);
				pn = strlen(*buffer);
				*(*buffer + pn) = 0;
				stop = FALSE;
			}
			break;
		default: break;
		}
		break;
	}
	if(NULL != rawdata)
		free(rawdata);
	if(NULL != payload)
		g_byte_array_free(payload, TRUE);
	purple_debug_misc("gotr", "%s========================\n", prompt);
	return stop;
}

static gboolean
chat_buddy_joining_cb(PurpleConversation *conv, const char *user,
		PurpleConvChatBuddyFlags flags, void *data)
{
	purple_debug_misc("gotr", "chat-buddy-joining (%s, %s, %d)\n",
			purple_conversation_get_name(conv), user, flags);

	return FALSE;
}

static void
chat_buddy_joined_cb(PurpleConversation *conv, const char *user,
		PurpleConvChatBuddyFlags flags, gboolean new_arrival, void *data)
{
	purple_debug_misc("gotr", "chat-buddy-joined (%s, %s, %d, %d)\n",
			purple_conversation_get_name(conv), user, flags, new_arrival);
}

static gboolean
chat_buddy_leaving_cb(PurpleConversation *conv, const char *user,
		const char *reason, void *data)
{
	purple_debug_misc("gotr", "chat-buddy-leaving (%s, %s, %s)\n",
			purple_conversation_get_name(conv), user, reason);

	return FALSE;
}

static void
chat_buddy_left_cb(PurpleConversation *conv, const char *user,
		const char *reason, void *data)
{
	purple_debug_misc("gotr", "chat-buddy-left (%s, %s, %s)\n",
			purple_conversation_get_name(conv), user, reason);
}

static gint
chat_invited_cb(PurpleAccount *account, const char *inviter,
		const char *room_name, const char *message,
		const GHashTable *components, void *data)
{
	purple_debug_misc("gotr", "chat-invited (%s, %s, %s, %s)\n",
			purple_account_get_username(account), inviter,
			room_name, message);

	return 0;
}

static void
chat_joined_cb(PurpleConversation *conv, void *data)
{
	GTree* toptable = (GTree*)data;
	PurpleConvChat* chat;
	gotr_t* chatdata;
	size_t nwritten;
	gpg_error_t err;

	purple_debug_misc("gotr", "chat-joined (%s)\n",
			purple_conversation_get_name(conv));

	if(NULL == toptable)	return;

	chat = purple_conversation_get_chat_data(conv);
	if(NULL == chat)		return; // not a chat

	chatdata = g_tree_lookup (toptable, GINT_TO_POINTER(chat->id));
	if(NULL != chatdata)	return; // keytable already exists

	chatdata = gotr_create_chatdata(chat->nick);
	if(NULL == chatdata)	return;

	// TODO: import this account's static key and store it to chatdata->tkey
	// TODO: other ways to get the prime and generator ??
	err = gcry_mpi_scan(&chatdata->prime, GCRYMPI_FMT_HEX, st_gotr_prime[GOTR_PRIMEINDEX], 0, &nwritten);
	if(err)	return;
	chatdata->g = mpi_set_ui(NULL, GOTR_G);

	g_tree_insert(toptable, GINT_TO_POINTER(chat->id), chatdata);
}

static void
chat_left_cb(PurpleConversation *conv, void *data)
{
	GTree* toptable = (GTree*)data;
	PurpleConvChat* chat;

	purple_debug_misc("gotr", "chat-left (%s)\n",
			purple_conversation_get_name(conv));

	if(NULL == toptable)	return;

	chat = purple_conversation_get_chat_data(conv);
	if(NULL == chat)		return; // not a chat

	g_tree_remove(toptable, GINT_TO_POINTER(chat->id));
}

/**************************************************************************
 * Plugin stuff
 **************************************************************************/
static gboolean
plugin_load(PurplePlugin *plugin)
{
	GTree* toptable = NULL;

	gotr_plugin_handle = plugin;

	/*  initializes the thread support subsystem in Libgcrypt
	err = gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_pthread);
	if(err) {
		purple_debug_misc ("gotr", "initializes the thread support subsystem in Libgcrypt failed\n");
		return FALSE;
	}
	 */
	/* Version check should be the very first call because it
	    makes sure that important subsystems are intialized. */
	if (!gcry_check_version (NULL))/*GCRYPT_VERSION*/
	{
		purple_debug_misc ("gotr", "libgcrypt version mismatch\n");
		return FALSE;
	}
	/* We don’t want to see any warnings, e.g. because we have not yet
	    parsed program options which might be used to suppress such
	    warnings. */
	gcry_control (GCRYCTL_SUSPEND_SECMEM_WARN);
	/* ... If required, other initialization goes here. Note that the
	    process might still be running with increased privileges and that
	    the secure memory has not been intialized. */
	/* Allocate a pool of secure memory. This make the secure memory
	    available and also drops privileges where needed. */
	gcry_control (GCRYCTL_INIT_SECMEM, 16384, 0);
	/* It is now okay to let Libgcrypt complain when there was/is
	   a problem with the secure memory. */
	gcry_control (GCRYCTL_RESUME_SECMEM_WARN);
	/* ... If required, other initialization goes here.  */
	/* Tell Libgcrypt that initialization has completed. */
	gcry_control (GCRYCTL_INITIALIZATION_FINISHED, 0);
	gcry_control (GCRYCTL_SET_DEBUG_FLAGS, 1u, 0);

	toptable = g_tree_new_full(gotr_int_cmp, NULL, NULL, gotr_free_chatdata) ;
	if(NULL == toptable) {
		purple_debug_misc("gotr", "creating top hash table failed!\n");
		return FALSE;
	}

	// Register a command to allow a user to enter /log some message and have
	// that message stored to the log. This command runs with default priority,
	// can only be used in a standard chat message, not while in group chat
	// mode
	keyupdate_command_id = purple_cmd_register
			(GOTR_KEYUPDATE_USERCMD,        /* command name */
					"",                            /* command argument format */
					PURPLE_CMD_P_DEFAULT,          /* command priority flags */
					PURPLE_CMD_FLAG_CHAT,          /* command usage flags */
					GOTR_PLUGIN_ID,                /* Plugin ID */
					gotr_keyupdate_cb,             /* Name of the callback function */
					GOTR_KEYUPDATE_HELP,           /* Help message */
					toptable                       /* Any special user-defined data */
			);

	core_handle     = purple_get_core();
	blist_handle    = purple_blist_get_handle();
	conn_handle     = purple_connections_get_handle();
	conv_handle     = purple_conversations_get_handle();
	accounts_handle = purple_accounts_get_handle();
	ciphers_handle  = purple_ciphers_get_handle();
	ft_handle       = purple_xfers_get_handle();
	sound_handle    = purple_sounds_get_handle();
	notify_handle   = purple_notify_get_handle();
	jabber_handle   = purple_plugins_find_with_id("prpl-jabber");

	purple_signal_connect(conv_handle, "sending-chat-msg",
			plugin, PURPLE_CALLBACK(sending_chat_msg_cb), toptable);
	purple_signal_connect(conv_handle, "receiving-chat-msg",
			plugin, PURPLE_CALLBACK(receiving_chat_msg_cb), toptable);

	purple_signal_connect(conv_handle, "chat-buddy-joining",
			plugin, PURPLE_CALLBACK(chat_buddy_joining_cb), toptable);
	purple_signal_connect(conv_handle, "chat-buddy-joined",
			plugin, PURPLE_CALLBACK(chat_buddy_joined_cb), toptable);
	purple_signal_connect(conv_handle, "chat-buddy-leaving",
			plugin, PURPLE_CALLBACK(chat_buddy_leaving_cb), toptable);
	purple_signal_connect(conv_handle, "chat-buddy-left",
			plugin, PURPLE_CALLBACK(chat_buddy_left_cb), toptable);

	purple_signal_connect(conv_handle, "chat-invited",
			plugin, PURPLE_CALLBACK(chat_invited_cb), toptable);
	purple_signal_connect(conv_handle, "chat-joined",
			plugin, PURPLE_CALLBACK(chat_joined_cb), toptable);
	purple_signal_connect(conv_handle, "chat-left",
			plugin, PURPLE_CALLBACK(chat_left_cb), toptable);

	return TRUE;
}

static gboolean
plugin_unload(PurplePlugin *plugin)
{

	// TODO: purple_signal_disconnect all signals
	// where to free rsc of toptable ??
	return TRUE;
}

static PurplePluginInfo info =
{
		PURPLE_PLUGIN_MAGIC,
		PURPLE_MAJOR_VERSION,
		PURPLE_MINOR_VERSION,
		PURPLE_PLUGIN_STANDARD,                             /**< type           */
		NULL,                                             /**< ui_requirement */
		0,                                                /**< flags          */
		NULL,                                             /**< dependencies   */
		PURPLE_PRIORITY_DEFAULT,                            /**< priority       */

		GOTR_PLUGIN_ID,                            /**< id             */
		"CrystalX",                               /**< name           */
		GOTR_PLUGIN_VERSION,                                  /**< version        */
		/**  summary        */
		"group otr.",
		/**  description    */
		"repudiable chat with origin authentication",
		"Hong Liu",       /**< author         */
		GOTR_PLUGIN_WEBSITE,                                     /**< homepage       */

		plugin_load,                                      /**< load           */
		plugin_unload,                                    /**< unload         */
		NULL,                                             /**< destroy        */

		NULL,                                             /**< ui_info        */
		NULL,                                             /**< extra_info     */
		NULL,
		NULL,
		/* Padding */
		NULL,
		NULL,
		NULL,
		NULL
};

static void
init_plugin(PurplePlugin *plugin)
{
}

PURPLE_INIT_PLUGIN(GOTR, init_plugin, info)

