/*
 * gotr_utility.h
 *
 *  Created on: Mar 26, 2012
 *      Author: lmm
 */

#ifndef GOTR_UTILITY_H_
#define GOTR_UTILITY_H_

#include <stdint.h>
#include <time.h>
#include <sys/times.h>
#include <glib.h>
#include <gpg-error.h>

#define GOTR_CIPHER (GCRY_CIPHER_AES)
//#define GOTR_HASH (GCRY_MD_SHA512)
#define GOTR_HASH (GCRY_MD_SHA256)

// in bits
// p=2q+1, p,q prime
extern const char* st_gotr_prime[2];
#define GOTR_2048
#ifdef  GOTR_2048
#define GOTR_SKEYSIZE (2048)
#define GOTR_PRIMEINDEX (0)
#else
#define GOTR_SKEYSIZE (4096)
#define GOTR_PRIMEINDEX (1)
#endif
#define GOTR_PRIMESIZE (GOTR_SKEYSIZE+1)
#define GOTR_PKEYSIZE GOTR_PRIMESIZE
#define GOTR_MAXPRPKEYLEN (GOTR_PRIMESIZE+31)
#define GOTR_G (4)

#define GOTR_MAXLINELEN (16384)
#define GOTR_PROMPTLEN  (100)
#define GOTR_SRV_QUEUESZ (100)
#define GOTR_ESC (0x7f)

// in bytes
#define GOTR_LEN_SIZE (4)
#define GOTR_BLOCKLEN (16)
#define GOTR_PADDING_MASK (~(GOTR_BLOCKLEN - 1UL))
// note that the following length is the max len, actual len may be less than this but still valid
#define GOTR_PRINTKENLEN (35)
#define GOTR_PLAINTXTLEN (GOTR_MAXLINELEN - GOTR_PRINTKENLEN)
#define GOTR_CIPHERTXTLEN (GOTR_PLAINTXTLEN + GOTR_PRINTKENLEN)

// restart pairwise init if the following timeout occurs
#define GOTR_INIT_PROBE_SLEEP (1)
#define GOTR_INIT_TIMEOUT (3)
#define GOTR_INIT_RETRY (8)
#define GOTR_UPDATE_PROBE_SLEEP (1)
#define GOTR_UPDATE_TIMEOUT (3)
#define GOTR_UPDATE_RETRY (8)
// the connection is closed if the following timeout occurs
#define GOTR_READ_TIMEOUT (3600)

//#define GOTR_VERBOSE
#ifdef GOTR_VERBOSE
#define gotr_printf(x...) printf(x)
#else
#define gotr_printf(x...)
#endif

/*
typedef enum {
	GOTR_OP_PAIR_INIT_E = 1234,
	GOTR_OP_PAIR_INIT_G,
	GOTR_OP_PAIR_INIT_X,
	GOTR_OP_PAIR_INIT_V,
	GOTR_OP_PAIR_UPDATE_G,
	GOTR_OP_PAIR_UPDATE_X,
	GOTR_OP_PAIR_UPDATE_V,
	GOTR_OP_MSG,
	GOTR_OP_NUM
}gotr_op_t;
*/
typedef enum {
	GOTR_STATE_IDLE = 0,
			GOTR_STATE_PAIR_INIT_S,
			GOTR_STATE_PAIR_INIT_R,
			GOTR_STATE_PAIR_INIT_G_S,
			GOTR_STATE_PAIR_INIT_G_R,
			GOTR_STATE_PAIR_INIT_X_S,
			GOTR_STATE_PAIR_INIT_X_R,
			GOTR_STATE_PAIR_INIT_V_S,
			GOTR_STATE_PAIR_INIT_V_R,
			GOTR_STATE_PAIR_UPDATE_G_S,
			GOTR_STATE_PAIR_UPDATE_G_R,
			GOTR_STATE_PAIR_UPDATE_X_S,
			GOTR_STATE_PAIR_UPDATE_X_R,
			GOTR_STATE_PAIR_UPDATE_V_S,
			GOTR_STATE_PAIR_UPDATE_V_R,
			GOTR_STATE_MSG,
			GOTR_STATE_NUM
}gotr_state_t;

typedef enum {
	GOTR_ERR_MEM = GPG_ERR_USER_1, // 1024
	GOTR_ERR_STAT, // 1025
	GOTR_ERR_STATE, // 1026
	GOTR_ERR_PRM, // 1027
	GOTR_ERR_SIG, // 1028
	GOTR_ERR_VRF, // 1029
	GOTR_ERR_CMD, // 102a
	GOTR_ERR_CRYPT, // 102b
	GOTR_ERR_OTHER // 102c
}gotr_err_t;

typedef enum {
	GOTR_KEY_STATIC_R,			// my static key
	GOTR_KEY_STATIC_PUB_R,	// my static pub key
	GOTR_KEY_STATIC_PUB_K,		// partner's static pub key
	GOTR_KEY_EPH_R,			// my ephemeral sig key
	GOTR_KEY_EPH_PUB_R,			// my pub ephemeral sig key
	GOTR_KEY_EPH_PUB_K, 			// partner's pub ephemeral sig key
	GOTR_KEY_R,
	GOTR_KEY_GR,
	GOTR_KEY_GK,
	GOTR_KEY_PRVPAIRR,
	GOTR_KEY_PUBPAIRR,
	GOTR_KEY_PUBPAIRK,
	GOTR_KEY_PAIRR,
	GOTR_KEY_FLAKE,
	GOTR_KEY_NEXT_EPH_R,			// my next ephemeral sig key
	GOTR_KEY_NEXT_EPH_PUB_R,
	GOTR_KEY_NEXT_EPH_PUB_K,
	GOTR_KEY_NEXT_R,
	GOTR_KEY_NEXT_GR,
	GOTR_KEY_NEXT_GK,
	GOTR_KEY_NEXT_PRVPAIRR,
	GOTR_KEY_NEXT_PUBPAIRR,
	GOTR_KEY_NEXT_PUBPAIRK,
	GOTR_KEY_NEXT_PAIRR,
	GOTR_KEY_NEXT_FLAKE
}gotr_keytype_t;

/**
 * we are going to organize gotr data in a hash map <int, gotr_t*>, where the first one is the id of a chat,
 * and pass it along with emitted messages
 */
typedef struct _gotr_keys_t {
	struct tms	execlk;			// exec time
	clock_t		etm;			// total elapsed time in time segments
	struct timeval tmval;		// total elapsed time acc to sys clock
	gcry_sexp_t pubkey; 		// partner's static pub key
	gcry_sexp_t sgr;			// my ephemeral sig key
	gcry_sexp_t sgk; 			// partner's pub ephemeral sig key
	gcry_mpi_t  r[2];			// my priv BD keys
	gcry_mpi_t  gr[2]; 			// my pub BD keys
	gcry_mpi_t  gk[2]; 			// partner's pub BD keys
	gcry_mpi_t  x0[2];			// my gadgets for the flake
	gcry_mpi_t  x1[2];			// my gadgets
	gcry_mpi_t  x2[2];			// partner's gadgets
	gcry_mpi_t  prvpairr;		// half pairwise sig key
	gcry_mpi_t  pubpairr;		// half pairwise sig key
	gcry_mpi_t  pubpairk;		// partner's half pairwise sig key
	gcry_mpi_t  pairrk;			// pairwise_BD_key
	gcry_mpi_t  flakek;			// flake_key
	uint32_t	noncer;
	uint32_t	noncek;
	gotr_state_t	state;		// state with the partner
	struct _gotr_keys_t * old;	// a list of history keys
	struct _gotr_keys_t * next;	// a partially negotiated key set
}gotr_keys_t;

typedef struct _gotr_t{
	GHashTable*				table;		// map of account_name to gotr_keys_t
	GTree*					revtable;	// map of mpi to Quark(account_name), the map is restrict to table[name], table[name]->old, and table[name]->new
	gcry_mpi_t 				g;
	gcry_mpi_t 				prime;
	gcry_sexp_t 			tkey; // my static asymm key
}gotr_t;

typedef GHashTableIter gotr_keys_iter_t;

gotr_keys_t* gotr_create_memdata(void);
void gotr_free_memdata(gpointer data);
void gotr_free(gpointer data);

gotr_t* gotr_create_chatdata(const char*);
void gotr_free_chatdata(gpointer data);

/**
 * check whether we have a record assoc with a name
 */
gboolean gotr_chatdata_lookupname(gotr_t* chatdata, const char* name);
/**
 * get internal static name string of a name
 * return value shon't be freed or modified
 */
const char* gotr_chatdata_getstaticname(gotr_t* chatdata, const char* name);

typedef struct _derived_id_t {
	const char*		name;
	gotr_keytype_t	type;
}derived_id_t;

/**
 * test whether a list of keys belong to the same acc and return the acc
 *  `name' in return value shon't be freed or modified
 */
derived_id_t* gotr_chatdata_deriveid(gotr_t* chatdata, GList* ls);

/**
 * get a list of valid gotr_keys_t (whose r[] gr[] gk[] x2[] are filled), resulting circle_key is stored to *pk if pk is not NULL
 */
GList* gotr_chatdata_getvalidkeyls(gotr_t* chatdata, gcry_mpi_t* pk);
/**
 * get the circle key from a list of pub BD keys and gadgets
 * @param 1: chat data
 * @param 2: ls of pubkeys
 * @param 3: ls of gadgets, they should be of equal length
 * @param 4: the derived identity of the sender
 * @return : the circle key, NULL if no pub BD keys are ours
 */
gcry_mpi_t gotr_chatdata_derivecirclek(gotr_t* chatdata, GList* lsk, GList* lsg, const char** hername);
/**
 * recalc flake key
 */
gcry_error_t gotr_calcflakekey(gotr_t* chatdata, gotr_keys_t* memdata);
void gotr_printflakekey(gotr_keys_t* handle);

gotr_keys_t* gotr_chatdata_getmap(gotr_t* chatdata, const char* name);
void gotr_chatdata_getmap_begin(gotr_t* chatdata, gotr_keys_iter_t* begin);	// a wrapper of g_hash_table_iter_init
int gotr_chatdata_getmap_next(gotr_keys_iter_t* iter, char** pname, gotr_keys_t** pmemdata);	// a wrapper of g_hash_table_iter_next
gboolean gotr_chatdata_addmap(gotr_t* chatdata, const char* name, gotr_keys_t* memdata); // this won't create a copy of memdata
gboolean gotr_chatdata_removemap(gotr_t* chatdata, const char* name);
/**
 * replace current key_set with the one `next` pointing to, and place `next` under `name`
 */
gboolean gotr_chatdata_updatemap(gotr_t* chatdata, const char* name);

const char* gotr_chatdata_getid(gotr_t* chatdata, const gcry_mpi_t key);
gboolean gotr_chatdata_addrevmap(gotr_t* chatdata, const gcry_mpi_t key, const char* name, gotr_keytype_t type);	// this will create a copy of key
gboolean gotr_chatdata_removerevmap(gotr_t* chatdata, const gcry_mpi_t key);

int gotr_compare_names(const char* a, const char* b) ;

/**************************************************************************
 * byte utilities
 **************************************************************************/

#define gotr_ceil(x) ((x + GOTR_BLOCKLEN - 1UL) & GOTR_PADDING_MASK)

/**
 * convert a string of bytes to its hex form
 * 	@return: newly allocated string, remember to free it after use
 */
char* gotr_printbytes(const unsigned char* pb, size_t len) ;

uint32_t read_ntohl(const void* buf);
void write_htonl(void* buf, uint32_t n);
uint16_t read_ntohs(const void* buf);
void write_htons(void* buf, uint16_t n);
size_t read_ntohll(const void* buf);
void write_htonll(void* buf, size_t n);

/**************************************************************************
 * mpi utilities
 **************************************************************************/
/**
 * generate a private BD key, and store it to *pp
 */
void gotr_genprivBDkey(gcry_mpi_t* pp);
/**
 * generate a private BD key, and store it to *ppgr
 */
void gotr_genpubBDkey (gcry_mpi_t* ppgr,
                    const gcry_mpi_t g, const gcry_mpi_t r,
                    const gcry_mpi_t m);
/**
 * generate a BD key pair, and store it to *ppr and *ppgr
 */
void gotr_genBDkeypair(gcry_mpi_t* ppr, gcry_mpi_t* ppgr, gotr_t* grp);
/**
 * generate a secret BD key, and store it to *ppgrk
 */
void gotr_genBDkey (gcry_mpi_t* ppgrk,
                    const gcry_mpi_t gk, const gcry_mpi_t r,
                    const gcry_mpi_t m);
/**
 * generate a gadget and store it to *ppx
 */
void gotr_gengadget (gcry_mpi_t* ppx,
                    const gcry_mpi_t gp1, const gcry_mpi_t gn1,
                    const gcry_mpi_t r,
                    const gcry_mpi_t m);

void gotr_free_mpi(gpointer key);
//gint gotr_compare_pubkeys(gconstpointer a, gconstpointer b) ;
gint gotr_compare_pubkeys(gconstpointer a, gconstpointer b, gpointer userdata);

/**
 * @param 1: iv, whose size is limited to GOTR_BLOCKLEN
 * @param 2: key
 * @param 3: data to be enc, no need to be padded; the output, however, will have rand padding, so it's ur job to remember original datalen
 * @param 4: data len, no need to be rounded
 * @param 5: out
 * @param 6: out len
 * @return: the addr to put in gcry_free
 */
unsigned char* gotr_enc(const unsigned char* iv, const gcry_mpi_t k, const unsigned char* data, size_t datalen, unsigned char** out, size_t* nwritten);
/**
 * @param 1: iv, whose size is limited to GOTR_BLOCKLEN
 * @param 2: key
 * @param 3: data to be dec, should be properly padded
 * @param 4: data len, should already be rounded
 * @param 5: out
 * @param 6: out len
 * @return: the addr to put in gcry_free
 */
unsigned char* gotr_dec(const unsigned char* iv, const gcry_mpi_t k, const unsigned char* data, size_t datalen, unsigned char** out, size_t* nwritten);
/**
 * @param 1: key, the sig is going to use hash(key)
 * @param 2: data to be enc, should already be padded
 * @param 3: data len
 * @param 4: out
 * @param 5: out len
 * @return: the addr to put in gcry_free
 */
unsigned char* gotr_sig(const gcry_mpi_t k, const unsigned char* data, size_t datalen, unsigned char** out);

void gotr_printmpi(const char* pre, gcry_mpi_t a, const char* post);
/**************************************************************************
 * gotr utilities
 **************************************************************************/

#define gotr_checkerr(x) if(x) break;

/**
 * encode encrypted message
 *  @param 1: input data stream
 * 	@param 2: length of the input data stream
 *	@param 3: output data stream, if is NULL a new output data stream will be created and returned
 * 	@return : output data stream that encodes input data stream
 * */
char* gotr_postproc_msg(const unsigned char* in, size_t len, char* ou);
/**
 * decode encrypted message
 * 	@param 1: input data stream
 * 	@param 2: output data stream, if is NULL a new output data stream will be created and returned
 *  @param 3: pointer to length of the output data stream
 * 	@return : output data stream that trans input data stream to binary data
 */
unsigned char* gotr_preproc_msg(const char* in, unsigned char* ou, size_t* plen);

/**
 *  an impl of g_list_free_full
 */
void gotr_list_free_full(GList *ls, GDestroyNotify free_func);
/**
 *  an impl of g_int_hash
 */
guint  gotr_int_hash(gconstpointer v);
/**
 *  an impl of g_int_equal
 */
gboolean gotr_int_equal (gconstpointer v1, gconstpointer v2);
/**
 *  an impl of g_int_equal
 */
gint gotr_int_cmp (gconstpointer v1, gconstpointer v2, gpointer user_data) ;
#endif /* GOTR_UTILITY_H_ */
